@@ -43,6 +43,7 @@ $ python setup.py help
 #error MAX_ML_REPLY must be 1 or 11 or 32
 #endif
 #define _FFR_MULTILINE (MAX_ML_REPLY > 1)
+#define PY_SSIZE_T_CLEAN
 
 //#include <pthread.h>	// shouldn't be needed - use Python API
 #include <Python.h>		// Python C API
@@ -250,7 +251,7 @@ static const char milter_set_flags__doc__[] =
 Set flags for filter capabilities; OR of one or more of:\n\
 ADDHDRS - filter may add headers\n\
 CHGBODY - filter may replace body\n\
-CHGFROM - filter may replace body\n\
+CHGFROM - filter may replace sender\n\
 ADDRCPT - filter may add recipients\n\
 DELRCPT - filter may delete recipients\n\
 CHGHDRS - filter may change/delete headers";
@@ -490,7 +491,7 @@ _generic_wrapper(milter_ContextObject *self, PyObject *cb, PyObject *arglist) {
   int retval;
 
   if (arglist == NULL) return _report_exception(self);
-  result = PyEval_CallObject(cb, arglist);
+  result = PyObject_CallObject(cb, arglist);
   Py_DECREF(arglist);
   if (result == NULL) return _report_exception(self);
 #if PY_MAJOR_VERSION >= 3
@@ -643,7 +644,7 @@ generic_env_wrapper(SMFICTX *ctx, PyObject*cb, char **argv) {
      /* There's some error checking performed in do_mkvalue() for a string */
      /* that's not currently done here - it probably should be */
 #if PY_MAJOR_VERSION >= 3
-     PyObject *o = PyUnicode_FromStringAndSize(argv[i], strlen(argv[i]));
+     PyObject *o = PyBytes_FromStringAndSize(argv[i], strlen(argv[i]));
 #else
      PyObject *o = PyString_FromStringAndSize(argv[i], strlen(argv[i]));
 #endif
@@ -674,7 +675,12 @@ milter_wrap_header(SMFICTX *ctx, char *headerf, char *headerv) {
    if (header_callback == NULL) return SMFIS_CONTINUE;
    c = _get_context(ctx);
    if (!c) return SMFIS_TEMPFAIL;
+#if PY_MAJOR_VERSION >= 3
+   /* pass val as bytes so Milter.Base.header_bytes can do surrogate escape. */
+   arglist = Py_BuildValue("(Osy)", c, headerf, headerv);
+#else
    arglist = Py_BuildValue("(Oss)", c, headerf, headerv);
+#endif
    return _generic_wrapper(c, header_callback, arglist);
 }
 
@@ -704,9 +710,9 @@ milter_wrap_body(SMFICTX *ctx, u_char *bodyp, size_t bodylen) {
    if (!c) return SMFIS_TEMPFAIL;
    /* Unclear whether this should be s#, z#, or t# */
 #if PY_MAJOR_VERSION >= 3
-   arglist = Py_BuildValue("(Oy#)", c, bodyp, bodylen);
+   arglist = Py_BuildValue("(Oy#)", c, bodyp, (Py_ssize_t)bodylen);
 #else
-   arglist = Py_BuildValue("(Os#)", c, bodyp, bodylen);
+   arglist = Py_BuildValue("(Os#)", c, bodyp, (Py_ssize_t)bodylen);
 #endif
    return _generic_wrapper(c, body_callback, arglist);
 }
@@ -898,7 +904,10 @@ milter_main(PyObject *self, PyObject *args) {
     return NULL;
   }
   /* libmilter requires thread support */
+  #if PY_VERSION_HEX < 0x03070000
+  /* called in Py_Initialize beginning with 3.7 */
   PyEval_InitThreads();	
+  #endif
   /* let other threads run while in smfi_main() */
   interp = PyThreadState_Get()->interp;
   _main = PyEval_SaveThread();	/* must be done before smfi_main() */
@@ -1129,6 +1138,8 @@ milter_chgfrom(PyObject *self, PyObject *args) {
   SMFICTX *ctx;
   PyThreadState *t;
   
+  /* FIXME: use s# to transition to allow passing bytes, but milter api
+   * requires NUL terminated bytes. */
   if (!PyArg_ParseTuple(args, "s|z:chgfrom", &sender, &params))
     return NULL;
   ctx = _find_context(self);
@@ -1225,7 +1236,7 @@ can only be called from the EOM callback.";
 static PyObject *
 milter_replacebody(PyObject *self, PyObject *args) {
   char *bodyp;
-  int bodylen;
+  Py_ssize_t bodylen;
   SMFICTX *ctx;
   PyThreadState *t;
   
@@ -1234,7 +1245,7 @@ milter_replacebody(PyObject *self, PyObject *args) {
   if (ctx == NULL) return NULL;
   t = PyEval_SaveThread();
   return _thread_return(t,smfi_replacebody(ctx,
-	(unsigned char *)bodyp, bodylen), "cannot replace message body");
+	(unsigned char *)bodyp, (int)bodylen), "cannot replace message body");
 }
 
 static const char milter_setpriv__doc__[] =
