@@ -3,6 +3,8 @@ import testmime
 import testsample
 import testutils
 import testgrey
+import testcfg
+import testpolicy
 import os
 
 def suite(): 
@@ -11,6 +13,8 @@ def suite():
   s.addTest(testsample.suite())
   s.addTest(testutils.suite())
   s.addTest(testgrey.suite())
+  s.addTest(testcfg.suite())
+  s.addTest(testpolicy.suite())
   return s
 
 if __name__ == '__main__':
